import six
from requests import Session

from restle.exceptions import HTTPException
from restle.serializers import URLSerializer


class Action(object):
    """Action base class"""

    NO_RESPONSE = 'none'
    DICT_RESPONSE = 'dict'
    OBJECT_RESPONSE = 'object'

    def __init__(self, relative_path, **kwargs):
        self.relative_path = relative_path
        self.http_method = kwargs.pop('http_method', 'POST')
        self.expected_http_codes = set(kwargs.pop('expected_http_code', six.moves.xrange(200, 300)))
        self.required_params = set(kwargs.pop('required_params', []))
        self.optional_params = set(kwargs.pop('optional_params', []))
        self.param_defaults = kwargs.pop('param_defaults', {})
        self.param_aliases = kwargs.pop('param_aliases', {})
        self.params_via_post = kwargs.pop('params_via_post', False)
        self.response_type = kwargs.pop('response_type', 'none')
        self.response_class = kwargs.pop('response_class', None)
        self.response_aliases = kwargs.pop('response_aliases', {})
        self.serializer = kwargs.pop('serializer', None)
        self.deserializer = kwargs.pop('deserializer', None)

        self.combined_params = self.optional_params.union(self.required_params)

        if kwargs:
            raise ValueError("Got unexpected keyword argument(s): '{0}'".format(', '.join(kwargs.keys())))

    def __call__(self, resource, **kwargs):
        invalid_kwargs = set(six.iterkeys(kwargs)).difference(self.combined_params)
        if invalid_kwargs:
            raise ValueError("Got unexpected keyword argument(s): '{0}'".format(', '.join(invalid_kwargs)))

        params = self.param_defaults.copy()
        params.update(kwargs)

        missing_required_params = self.required_params.difference(set(six.iterkeys(params)))
        if missing_required_params:
            raise ValueError("Missing required parameter(s): '{0}'".format(', '.join(missing_required_params)))

        params, content_type = self.prepare_params({self.param_aliases.get(k, k): v for k, v in six.iteritems(params)})
        return self.process_response(
            self.do_request(self.get_uri(resource._url), params, content_type, resource._session)
        )

    def contribute_to_class(self, cls, name):
        self._attr_name = name
        self._resource = cls

        cls._meta.actions.append(self)

    def get_uri(self, base_uri):
        if not base_uri.endswith('/') and not self.relative_path.startswith('/'):
            base_uri += '/'

        return ''.join((base_uri, self.relative_path))

    def prepare_params(self, params):
        if self.serializer:
            serializer = self.serializer
        elif self.params_via_post and self.http_method in ('POST', 'PUT', 'PATCH'):
            serializer = self._resource._meta.serializer
        else:
            serializer = URLSerializer

        return serializer.to_string(params), serializer.content_type

    def do_request(self, url, params, content_type, session=None):
        body = None
        headers = None
        if params and not self.params_via_post:
            url += '?{0}'.format(params)
        elif self.params_via_post:
            body = params
            headers = {'Content-type': content_type}

        if session is None:
            session = Session()

        return getattr(session, self.http_method.lower())(url, data=body, headers=headers)

    def process_response(self, response):
        if response.status_code not in self.expected_http_codes:
            raise HTTPException(
                'Received unexpected response from server: {0} ({1})'.format(response.status_code, response.reason)
            )

        if self.response_type == self.NO_RESPONSE:
            return

        data = (self.deserializer or self._resource._meta.serializer).to_dict(response.text)

        def alias_keys(d):
            if isinstance(d, dict):
                return {self.response_aliases.get(k, k): alias_keys(v) for k, v in six.iteritems(d)}
            elif isinstance(d, list):
                return [alias_keys(x) for x in d]
            return d

        data = alias_keys(data)

        if self.response_type == self.OBJECT_RESPONSE:
            if self.response_class:
                return self.response_class(data)
            else:
                def convert_to_class(d):
                    if isinstance(d, dict):
                        return type('AnonymousObject', (), {k: convert_to_class(v) for k, v in six.iteritems(d)})
                    elif isinstance(d, list):
                        return [convert_to_class(x) for x in d]
                    return d

                return convert_to_class(data)
        return data



import copy
import logging
import string

import six
from requests import Session

from restle.exceptions import NotFoundException, HTTPException, MissingFieldException
from restle.options import ResourceOptions

logger = logging.getLogger(__name__)

ALPHANUMERIC = set(string.ascii_letters + string.digits)


class ResourceBase(type):
    """Resource metaclass"""

    def __new__(cls, name, bases, attrs):
        super_new = super(ResourceBase, cls).__new__
        module = attrs.pop('__module__', None)

        new_attrs = {'__module__': module}
        classcell = attrs.pop('__classcell__', None)
        if classcell is not None:
            new_attrs['__classcell__'] = classcell

        new_class = super_new(cls, name, bases, new_attrs)
        meta = attrs.pop('Meta', None)
        new_class.add_to_class('_meta', ResourceOptions(meta))

        for name, value in attrs.items():
            new_class.add_to_class(name, value)

        field_names = set(x.name for x in new_class._meta.fields)

        for base in bases:
            base_fields = [x for x in getattr(getattr(base, '_meta', None), 'fields', []) if x.name not in field_names]
            new_class._meta.fields = base_fields + new_class._meta.fields

        return new_class

    def add_to_class(cls, name, value):
        if hasattr(value, 'contribute_to_class'):
            value.contribute_to_class(cls, name)
        else:
            setattr(cls, name, value)


class Resource(six.with_metaclass(ResourceBase)):
    def __init__(self, **kwargs):
        self._session = kwargs.pop('session', None)
        self._url = None
        self._params = None
        self._strict = True
        self._populated_field_values = True if kwargs else False

        if self._session is None:
            self._session = Session()

        for field in self._meta.fields:
            if field._attr_name in kwargs:
                setattr(self, field._attr_name, kwargs.pop(field._attr_name))

        def action_wrapper(fn):
            def inner(*args, **kwargs):
                return fn(self, *args, **kwargs)
            return inner

        for action in self._meta.actions:
            setattr(self, action._attr_name, action_wrapper(action))

        if kwargs:
            raise TypeError('Resource received invalid keyword argument(s): {0}'.format(', '.join(kwargs.keys())))

    def _load_resource(self):
        """Load resource data from server"""

        url = self._url
        if self._params:
            url += '?{0}'.format(six.moves.urllib_parse.urlencode(self._params))

        r = getattr(self._session, self._meta.get_method.lower())(url)

        if r.status_code == 404:
            raise NotFoundException('Server returned 404 Not Found for the URL {0}'.format(self._url))
        elif not 200 <= r.status_code < 400:
            raise HTTPException('Server returned {0} ({1})'.format(r.status_code, r.reason), r)

        data = self._meta.deserializer.to_dict(r.text)
        self.populate_field_values(data)

    def populate_field_values(self, data):
        """Load resource data and populate field values"""

        if not self._meta.case_sensitive_fields:
            data = {k.lower(): v for k, v in six.iteritems(data)}

        if self._meta.match_fuzzy_keys:
            # String any non-alphanumeric chars from each key
            data = {''.join(x for x in k if x in ALPHANUMERIC).lower(): v for k, v in six.iteritems(data)}

        for field in self._meta.fields:
            name = field.name if self._meta.case_sensitive_fields else field.name.lower()
            value = None

            if self._meta.match_fuzzy_keys:
                name = ''.join(x for x in name if x in ALPHANUMERIC).lower()

            if name in data:
                value = field.to_python(data[name], self)
            elif field.required and field.default is None:
                message = "Response from {0} is missing required field '{1}'".format(self._url, field.name)
                if self._strict:
                    raise MissingFieldException(message)
                else:
                    logger.warn(message)
            elif field.default is not None:
                value = copy.copy(field.default)

            setattr(self, field._attr_name, value)

        self._populated_field_values = True

    def __getattr__(self, item):
        if self._populated_field_values:
            raise AttributeError("'{0}' object has no attribute '{1}'".format(self.__class__.__name__, item))

        self._load_resource()
        return getattr(self, item)

    @classmethod
    def get(cls, url, strict=True, lazy=True, session=None):
        self = cls(session=session)
        o = six.moves.urllib_parse.urlparse(url)

        self._params = self._meta.get_parameters.copy()
        if o.query:
            query = six.moves.urllib_parse.parse_qs(o.query)
            self._params.update(
                {k: v[0] if v else '' for k, v in six.iteritems(query)}
            )

        self._url = '{0}://{1}{2}'.format('https' if self._meta.force_https else o.scheme, o.netloc, o.path)
        self._strict = strict

        if not lazy:
            self._load_resource()

        return self



from restle.serializers import JSONSerializer, URLSerializer

OPTION_NAMES = (
    'case_sensitive_fields', 'match_fuzzy_keys', 'force_https', 'get_method', 'get_parameters', 'deserializer',
    'serializer'
)


class ResourceOptions(object):
    def __init__(self, meta):
        self.case_sensitive_fields = True
        self.match_fuzzy_keys = False
        self.force_https = False
        self.get_method = 'GET'
        self.get_parameters = {}
        self.deserializer = JSONSerializer()
        self.serializer = URLSerializer()

        self.fields = []
        self.actions = []
        self.meta = meta

    def contribute_to_class(self, cls, name):
        cls._meta = self

        if self.meta:
            meta_attrs = self.meta.__dict__.copy()

            # Remove private attributes
            for key in self.meta.__dict__:
                if key.startswith('_'):
                    del meta_attrs[key]

            for name in OPTION_NAMES:
                if name in meta_attrs:
                    setattr(self, name, meta_attrs.pop(name))

            # Check for invalid attributes
            if meta_attrs:
                raise TypeError('Meta class contains invalid attribute(s): {0}'.format(', '.join(meta_attrs.keys())))

            del self.meta



import six


class ResourceException(Exception):
    pass


class NotFoundException(ResourceException):
    pass


class MissingFieldException(ResourceException):
    pass


class HTTPException(six.moves.http_client.HTTPException):
    def __init__(self, message, response=None):
        super(HTTPException, self).__init__(message)
        self.response = response



import json
import six


class JSONSerializer(object):
    content_type = 'application/json'

    @staticmethod
    def to_dict(s):
        return json.loads(s, strict=False)

    @staticmethod
    def to_string(d):
        return json.dumps(d)


class URLSerializer(object):
    content_type = 'application/x-www-form-urlencoded'

    @staticmethod
    def to_dict(s):
        return six.moves.urllib_parse.parse_qs(s)

    @staticmethod
    def to_string(d):
        return six.moves.urllib_parse.urlencode(d)


import subprocess
import sys
from setuptools import setup, Command

import restle


class PyTest(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        errno = subprocess.call([sys.executable, 'runtests.py'])
        raise SystemExit(errno)

setup(
    name='restle',
    description='A REST client framework',
    keywords='rest,mapper,client',
    version=restle.__version__,
    packages=['restle'],
    requires=['six', 'requests'],
    url='https://github.com/consbio/restle',
    license='BSD',
    tests_require=['pytest', 'httpretty==0.8.6', 'mock'],
    cmdclass={'test': PyTest}
)



import six


class Field(object):
    """Field base class"""

    def __init__(self, name=None, required=True, default=None):
        self._attr_name = None
        self._resource = None

        self.name = name
        self.required = required
        self.default = default

    def contribute_to_class(self, cls, name):
        self._attr_name = name
        self._resource = cls

        if not self.name:
            self.name = name

        cls._meta.fields.append(self)

    def to_python(self, value, resource):
        """Returns the value as returned by the serializer converted to a Python object"""

        return value

    def to_value(self, obj, resource):
        """Returns the Python object converted to a value ready for serialization"""

        return obj


class DictField(Field):
    """Same as base field. Aliased for semantic purposes."""

    pass


class ListField(Field):
    def to_python(self, value, resource):
        if not isinstance(value, (list, tuple, set)):
            raise ValueError("Expected a list, got '{0}'".format(value.__class__.__name__))

        return list(value)


class TextField(Field):
    def __init__(self, encoding='utf-8', strip=False, lower=False, *args, **kwargs):
        """
        :param encoding: Assume this encoding for incoming values
        :param strip: If True, will remove leading and trailing whitespace from value
        :param lower: If True, will convert value to lower case
        """

        super(TextField, self).__init__(*args, **kwargs)

        self.encoding = encoding
        self.strip = strip
        self.lower = lower

    def _transform(self, value):
        if self.strip:
            value = value.strip()

        if self.lower:
            value = value.lower()

        return value

    def to_python(self, value, resource):
        """Converts to unicode if `self.encoding != None`, otherwise returns input without attempting to decode"""

        if value is None:
            return self._transform(value)

        if isinstance(value, six.text_type):
            return self._transform(value)

        if self.encoding is None and isinstance(value, (six.text_type, six.binary_type)):
            return self._transform(value)

        if self.encoding is not None and isinstance(value, six.binary_type):
            return self._transform(value.decode(self.encoding))

        return self._transform(six.text_type(value))


class BooleanField(Field):
    def to_python(self, value, resource):
        if value is None:
            return value

        return bool(value)


class NumberField(Field):
    def to_python(self, value, resource):
        if isinstance(value, (int, float)) or value is None:
            return value

        number = float(value)
        return int(number) if number.is_integer() else number


class IntegerField(NumberField):
    def to_python(self, value, resource):
        if isinstance(value, int) or value is None:
            return value

        return int(super(IntegerField, self).to_python(value, resource))

    def to_value(self, obj, resource):
        if obj is None:
            return obj

        return int(obj)


class FloatField(NumberField):
    def to_python(self, value, resource):
        if value is None:
            return value

        return float(super(FloatField, self).to_python(value, resource))

    def to_value(self, obj, resource):
        if obj is None:
            return obj

        return float(obj)


class ObjectField(Field):
    """Represents a dictionary as a Python object (lists, too)"""

    def __init__(self, class_name='AnonymousObject', aliases={}, *args, **kwargs):
        self.class_name = class_name
        self.aliases = aliases
        self.reverse_aliases = {v: k for k, v in six.iteritems(aliases)}

        super(ObjectField, self).__init__(*args, **kwargs)

    def to_python(self, value, resource):
        """Dictionary to Python object"""

        if isinstance(value, dict):
            d = {
                self.aliases.get(k, k): self.to_python(v, resource) if isinstance(v, (dict, list)) else v
                for k, v in six.iteritems(value)
            }
            return type(self.class_name, (), d)
        elif isinstance(value, list):
            return [self.to_python(x, resource) if isinstance(x, (dict, list)) else x for x in value]
        else:
            return value

    def to_value(self, obj, resource, visited=set()):
        """Python object to dictionary"""

        if id(obj) in visited:
            raise ValueError('Circular reference detected when attempting to serialize object')

        if isinstance(obj, (list, tuple, set)):
            return [self.to_value(x, resource) if hasattr(x, '__dict__') else x for x in obj]
        elif hasattr(obj, '__dict__'):
            attrs = obj.__dict__.copy()
            for key in six.iterkeys(obj.__dict__):
                if key.startswith('_'):
                    del attrs[key]

            return {
                self.reverse_aliases.get(k, k):
                    self.to_value(v, resource) if hasattr(v, '__dict__') or isinstance(v, (list, tuple, set)) else v
                for k, v in six.iteritems(attrs)
            }
        else:
            return obj


class NestedResourceField(Field):
    """Base class for nested resource fields"""

    ID_ONLY = 'id'
    PARTIAL_OBJECT = 'partial'
    FULL_OBJECT = 'full'

    def __init__(self, resource_class, nest_type, id_field='id', relative_path=None, *args, **kwargs):
        """
        :param str nest_type: One of 'id', 'partial', 'full' depending on whether the resource is expanded or needs to
        be loaded separately.
        :param id_field: For types 'id' and 'partial', specifies which field will be used as the nested resource id
        when constructing the URI.
        :param relative_path: The relative path (from this resource) to the nested resource. May contain {id} which
        will be replaced with the resource id. E.g. '/nested-resource/{id}/'
        """

        super(NestedResourceField, self).__init__(*args, **kwargs)

        if nest_type == self.ID_ONLY and not relative_path:
            raise ValueError("Nested resources of type 'uri' must provide a relative_path argument.")

        if nest_type == self.PARTIAL_OBJECT and not (id_field and relative_path):
            raise ValueError("Nested resources of type 'partial' must specify 'id_field' and 'relative_path'")

        self.resource_class = resource_class
        self.type = nest_type
        self.id_field = id_field
        self.relative_path = relative_path

    def get_uri(self, obj, base_uri):
        if not base_uri.endswith('/') and not self.relative_path.startswith('/'):
            base_uri += '/'

        if self.type == self.ID_ONLY:
            resource_id = obj
        else:
            resource_id = obj.get(self.id_field)

        return ''.join((base_uri, self.relative_path.format(id=resource_id)))

    def to_python(self, value, resource):
        if value is None:
            return value

        if self.type in (self.PARTIAL_OBJECT, self.FULL_OBJECT) and not isinstance(value, dict):
            raise ValueError(
                "Expected nested resource to be of type 'dict', got '{0}'".format(value.__class__.__name__)
            )
        elif self.type == self.ID_ONLY and not isinstance(value, (six.string_types, int)):
            raise ValueError(
                "Expected nested resource to be a string or int, got type {0}'".format(value.__class__.__name__)
            )

        if self.type == self.FULL_OBJECT:
            nested = self.resource_class()
            nested.populate_field_values(value)

            if self.relative_path:
                nested._url = self.get_uri(value, resource._url)

            return nested
        else:
            return self.resource_class.get(self.get_uri(value, resource._url), session=resource._session)

    def to_value(self, obj, resource):
        raise NotImplementedError('Serializing nested resources is not yet supported')


class ToOneField(NestedResourceField):
    """Same as NestedResourceField. Aliased for semantic reasons"""

    pass


class ToManyField(NestedResourceField):
    """To-many nested resource field"""

    def __iter__(self):
        """Implementing __iter__ avoids IDE inspection errors/warnings when this field is used in iteration"""

        while False:
            yield None

    def to_python(self, value, resource):
        if value is None:
            return []

        if not isinstance(value, list):
            raise ValueError("Expected a list for 'to many' value, got '{0}'".format(value.__class__.__name__))

        return [super(ToManyField, self).to_python(x, resource) for x in value]

    def to_value(self, obj, resource):
        raise NotImplementedError('Serializing nested resources is not yet supported')
